/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package moe.nekojimi.nodeprocessor.links;

import moe.nekojimi.nodeprocessor.nodes.Node;
import moe.nekojimi.nodeprocessor.ports.ConstantOutPort;
import moe.nekojimi.nodeprocessor.ports.InPort;
import moe.nekojimi.nodeprocessor.ports.OutPort;
import moe.nekojimi.nodeprocessor.ports.QueueOutPort;

/**
 *
 * @author jimj316
 */
public abstract class Link<T, I extends InPort<T>, O extends OutPort<T>>
{
    final O from;
    private final I to;

    public Link(O from, I to)
    {
        this.from = from;
        this.to = to;
        connect();
    }

    public final void connect()
    {
        from.connect(this);
        to.connect(this);
        from.getNode().getProgram().addLink(this);
        to.getNode().getProgram().addLink(this);
    }

    public final void disconnect()
    {
        from.disconnect(this);
        to.disconnect(this);
    }

    public abstract T read();

    public abstract void write(T val);

    public abstract boolean has();

    public static <T> Link<T, ?, ?> makeLink(Class<T> clazz, Node from, String fromPort, Node to, String toPort)
    {
        OutPort<T> outPort = from.getOutPort(fromPort, clazz);
        InPort<T> inPort = to.getInPort(toPort, clazz);
        if (outPort instanceof QueueOutPort)
        {
            QueueOutPort<T> q = (QueueOutPort<T>) outPort;
            return new QueueLink<>(q, inPort);
        } else if (outPort instanceof ConstantOutPort)
        {
            ConstantOutPort<T> c = (ConstantOutPort<T>) outPort;
            return new ConstantLink<>(c, inPort);
        } else
        {
            return null;
        }
    }

    protected void signal()
    {
        to.signal();
    }

    public O getFrom()
    {
        return from;
    }

    public I getTo()
    {
        return to;
    }

}
