/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package moe.nekojimi.nodeprocessor.nodes;

import moe.nekojimi.nodeprocessor.ports.InPort;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Parameter;
import java.util.logging.Level;
import java.util.logging.Logger;
import moe.nekojimi.nodeprocessor.Program;
import moe.nekojimi.nodeprocessor.ports.QueueOutPort;

/**
 *
 * @author jimj316
 * @param <C> the class that declares this method.
 */
public class MethodNode<C> extends Node
{

    private final InPort<C> objectInPort;
    private final QueueOutPort returnOutPort;
    private final Method method;

    public MethodNode(Program scene, Class<C> clazz, Method method)
    {
        super(scene);
        assert (method.getDeclaringClass() == clazz);
        if (Modifier.isStatic(method.getModifiers()))
            objectInPort = null;
        else
        {
            objectInPort = new InPort<>("object", clazz, this);
            addInPort(objectInPort);
        }
        this.method = method;
        Parameter[] parameters = method.getParameters();
        for (Parameter parameter : parameters)
            addInPort(new InPort(parameter.getName(), parameter.getType(), this));
        Class<?> returnType = method.getReturnType();
        returnOutPort = new QueueOutPort("out", returnType, this);
        addOutPort(returnOutPort);

        setName(method.getDeclaringClass().getSimpleName() + " " + method.getName());
    }


    @Override
    public boolean process()
    {
        try
        {
            C object = null;
            if (objectInPort != null)
                object = objectInPort.get();
            Object ret = method.invoke(object);
            returnOutPort.set(ret);
        } catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException ex)
        {
            Logger.getLogger(MethodNode.class.getName()).log(Level.SEVERE, null, ex);
        }
        return true;
    }

    @Override
    public boolean check()
    {
        return getInPorts().stream().allMatch(in -> (in.has()));
    }

}
