/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package moe.nekojimi.nodeprocessor.nodes;

import java.util.*;
import moe.nekojimi.nodeprocessor.ports.InPort;
import moe.nekojimi.nodeprocessor.ports.OutPort;
import moe.nekojimi.nodeprocessor.Program;

/**
 *
 * @author jimj316
 */
public abstract class Node
{
//    protected Map<String, Link> inLinks = new HashMap<>();
//    protected Map<String, Link> outLinks = new HashMap<>();

    private final Program program;
    private Map<String, InPort> inPorts = new HashMap<>();
    private Map<String, OutPort> outPorts = new HashMap<>();
    private String name;

    public Node(Program program)
    {
        this.program = program;
        program.addNode(this);
        this.name = this.getClass().getSimpleName();
    }

    public String getName()
    {
        return name;
    }

    protected void setName(String name)
    {
        this.name = name;
    }

    protected final void addInPort(InPort port)
    {
        inPorts.putIfAbsent(getPortName(port.getClazz(), port.getName()), port);
    }
    protected final void addOutPort(OutPort port)
    {
        outPorts.putIfAbsent(getPortName(port.getClazz(), port.getName()), port);
    }

    public final <X> InPort<X> getInPort(String name, Class<X> clazz)
    {
        Class<? super X> c = clazz;
        InPort ret;
        do
        {
            ret = inPorts.get(getPortName(c, name));
            c = c.getSuperclass();
        } while (ret == null && c != null);
        return ret;
    }

    public final <X> OutPort<X> getOutPort(String name, Class<X> clazz)
    {
        Class<? super X> c = clazz;
        OutPort ret;
        do
        {
            ret = outPorts.get(getPortName(clazz, name));
            c = c.getSuperclass();
        } while (ret == null && c != null);
        return ret;
    }

    public final Collection<InPort> getInPorts()
    {
        return inPorts.values();
    }

    public final Collection<OutPort> getOutPorts()
    {
        return outPorts.values();
    }

    private static <X> String getPortName(Class<X> clazz, String name)
    {
        return clazz.getCanonicalName() + "-" + name;
    }

    public int getNumberOfInputs()
    {
        return inPorts.values().stream().mapToInt((t) -> t.getNumberOfConnections()).sum();
    }

    public int getNumberOfOutputs()
    {
        return outPorts.values().stream().mapToInt((t) -> t.getNumberOfConnections()).sum();
    }

    public Program getProgram()
    {
        return program;
    }

    public boolean dependsOn(Node other)
    {
        // TODO: this is a crude search that will enter an infinite loop if a node depends on itself
        for (InPort port : inPorts.values())
        {
            Set<Node> linkedNodes = port.getLinkedNodes();
            if (linkedNodes.contains(other))
                return true;
            else
                for (Node dependent : linkedNodes)
                {
                    if (dependent.dependsOn(other))
                        return true;
                }
        }
        return false;
    }

    /**
     * Processes this node's inputs, and generates output (or performs some
     * other useful function).
     *
     * @return true if there will be no more output until an input is received;
     * false if process() should be called again.
     */
    public abstract boolean process();

    public abstract boolean check();

    public void signal()
    {
        if (check())
            program.queueNodeProcess(this);
    }

    public void reset()
    {
        //  assume most nodes are stateless
        // default implementation does nothing
    }

    public static final Comparator<Node> DEPENDENCY_ORDER = (Node a, Node b) ->
    {
        int ret = 0;
        if (a.dependsOn(b))
            ret--;
        if (b.dependsOn(a))
            ret++;
        return ret;
    };

}
